@@ -1,4 +1,5 @@
 """Tools for downloading map tiles from coordinates."""
+
 from __future__ import absolute_import, division, print_function
 
 import uuid
@@ -7,7 +8,7 @@ import mercantile as mt
 import requests
 import atexit
 import io
-import os
+import time
 import shutil
 import tempfile
 import warnings
@@ -72,6 +73,7 @@ def bounds2raster(
     path,
     zoom="auto",
     source=None,
+    headers: dict[str, str] | None = None,
     ll=False,
     wait=0,
     max_retries=2,
@@ -105,6 +107,9 @@ def bounds2raster(
         `rasterio` and all bands are loaded into the basemap.
         IMPORTANT: tiles are assumed to be in the Spherical Mercator
         projection (EPSG:3857), unless the `crs` keyword is specified.
+    headers : dict[str, str] or None
+        [Optional. Default: None]
+        Headers to include with requests to the tile server.
     ll : Boolean
         [Optional. Default: False] If True, `w`, `s`, `e`, `n` are
         assumed to be lon/lat as opposed to Spherical Mercator.
@@ -135,13 +140,26 @@ def bounds2raster(
     extent : tuple
         Bounding box [minX, maxX, minY, maxY] of the returned image
     """
+    if headers is None:
+        headers = {}
+
     if not ll:
         # Convert w, s, e, n into lon/lat
         w, s = _sm2ll(w, s)
         e, n = _sm2ll(e, n)
     # Download
-    Z, ext = bounds2img(w, s, e, n, zoom=zoom, source=source, ll=True, n_connections=n_connections,
-                        use_cache=use_cache)
+    Z, ext = bounds2img(
+        w,
+        s,
+        e,
+        n,
+        zoom=zoom,
+        source=source,
+        headers=headers,
+        ll=True,
+        n_connections=n_connections,
+        use_cache=use_cache,
+    )
 
     # Write
     # ---
@@ -171,7 +189,19 @@ def bounds2raster(
 
 
 def bounds2img(
-    w, s, e, n, zoom="auto", source=None, ll=False, wait=0, max_retries=2, n_connections=1, use_cache=True, zoom_adjust=None
+    w,
+    s,
+    e,
+    n,
+    zoom="auto",
+    source=None,
+    headers: dict[str, str] | None = None,
+    ll=False,
+    wait=0,
+    max_retries=2,
+    n_connections=1,
+    use_cache=True,
+    zoom_adjust=None,
 ):
     """
     Take bounding box and zoom and return an image with all the tiles
@@ -198,6 +228,9 @@ def bounds2img(
         `rasterio` and all bands are loaded into the basemap.
         IMPORTANT: tiles are assumed to be in the Spherical Mercator
         projection (EPSG:3857), unless the `crs` keyword is specified.
+    headers : dict[str, str] or None
+        [Optional. Default: None]
+        Headers to include with requests to the tile server.
     ll : Boolean
         [Optional. Default: False] If True, `w`, `s`, `e`, `n` are
         assumed to be lon/lat as opposed to Spherical Mercator.
@@ -232,6 +265,9 @@ def bounds2img(
     extent : tuple
         Bounding box [minX, maxX, minY, maxY] of the returned image
     """
+    if headers is None:
+        headers = {}
+
     if not ll:
         # Convert w, s, e, n into lon/lat
         w, s = _sm2ll(w, s)
@@ -251,16 +287,17 @@ def bounds2img(
     tile_urls = [provider.build_url(x=tile.x, y=tile.y, z=tile.z) for tile in tiles]
     # download tiles
     if n_connections < 1 or not isinstance(n_connections, int):
-        raise ValueError(
-            f"n_connections must be a positive integer value."
-        )
+        raise ValueError(f"n_connections must be a positive integer value.")
     # Use threads for a single connection to avoid the overhead of spawning a process. Use processes for multiple
     # connections if caching is enabled, as threads lead to memory issues when used in combination with the joblib
     # memory caching (used for the _fetch_tile() function).
-    preferred_backend = "threads" if (n_connections == 1 or not use_cache) else "processes"
+    preferred_backend = (
+        "threads" if (n_connections == 1 or not use_cache) else "processes"
+    )
     fetch_tile_fn = memory.cache(_fetch_tile) if use_cache else _fetch_tile
     arrays = Parallel(n_jobs=n_connections, prefer=preferred_backend)(
-        delayed(fetch_tile_fn)(tile_url, wait, max_retries) for tile_url in tile_urls)
+        delayed(fetch_tile_fn)(tile_url, wait, max_retries, headers) for tile_url in tile_urls
+    )
     # merge downloaded tiles
     merged, extent = _merge_tiles(tiles, arrays)
     # lon/lat extent --> Spheric Mercator
@@ -287,8 +324,8 @@ def _process_source(source):
     return provider
 
 
-def _fetch_tile(tile_url, wait, max_retries):
-    array = _retryer(tile_url, wait, max_retries)
+def _fetch_tile(tile_url, wait, max_retries, headers: dict[str, str]):
+    array = _retryer(tile_url, wait, max_retries, headers)
     return array
 
 
@@ -406,7 +443,7 @@ def _warper(img, transform, s_crs, t_crs, resampling):
     return img, bounds, transform
 
 
-def _retryer(tile_url, wait, max_retries):
+def _retryer(tile_url, wait, max_retries, headers: dict[str, str]):
     """
     Retry a url many times in attempt to get a tile and read the image
 
@@ -421,13 +458,15 @@ def _retryer(tile_url, wait, max_retries):
     max_retries : int
         total number of rejected requests allowed before contextily
         will stop trying to fetch more tiles from a rate-limited API.
+    headers: dict[str, str]
+        headers to include with request.
 
     Returns
     -------
     array of the tile
     """
     try:
-        request = requests.get(tile_url, headers={"user-agent": USER_AGENT})
+        request = requests.get(tile_url, headers={"user-agent": USER_AGENT, **headers})
         request.raise_for_status()
         with io.BytesIO(request.content) as image_stream:
             image = Image.open(image_stream).convert("RGBA")
@@ -442,15 +481,15 @@ def _retryer(tile_url, wait, max_retries):
                 "Tile URL resulted in a 404 error. "
                 "Double-check your tile url:\n{}".format(tile_url)
             )
-        elif request.status_code == 104 or request.status_code == 200:
+        else:
             if max_retries > 0:
-                os.wait(wait)
+                time.sleep(wait)
                 max_retries -= 1
-                request = _retryer(tile_url, wait, max_retries)
+                request = _retryer(tile_url, wait, max_retries, headers)
             else:
-                raise requests.HTTPError("Connection reset by peer too many times.")
-
-
+                raise requests.HTTPError("Connection reset by peer too many times. "
+                                         f"Last message was: {request.status_code} "
+                                         f"Error: {request.reason} for url: {request.url}")
 
 def howmany(w, s, e, n, zoom, verbose=True, ll=False):
     """
@@ -545,7 +584,7 @@ def _sm2ll(x, y):
 def _calculate_zoom(w, s, e, n):
     """Automatically choose a zoom level given a desired number of tiles.
 
-    .. note:: all values are interpreted as latitude / longitutde.
+    .. note:: all values are interpreted as latitude / longitude.
 
     Parameters
     ----------
